from ytmusicapi.navigation import *
from ytmusicapi.type_alias import JsonList


def parse_artists_runs(runs: JsonList) -> JsonList:
    """Returns artist names and IDs. Skips every other run to avoid separators."""
    artists = []
    for j in range(int(len(runs) / 2) + 1):
        artists.append({"name": runs[j * 2]["text"], "id": nav(runs[j * 2], NAVIGATION_BROWSE_ID, True)})
    return artists



from urllib.parse import parse_qs, urlparse

from ytmusicapi.continuations import *
from ytmusicapi.enums import ResponseStatus
from ytmusicapi.exceptions import YTMusicUserError
from ytmusicapi.helpers import sum_total_duration
from ytmusicapi.models.content.enums import PlaylistSortOrder, PlaylistVoteEditOptions
from ytmusicapi.navigation import *
from ytmusicapi.parsers.browsing import parse_content_list, parse_playlist
from ytmusicapi.parsers.playlists import *
from ytmusicapi.type_alias import JsonDict, JsonList, ParseFuncType, RequestFuncBodyType, RequestFuncType

from ._protocol import MixinProtocol
from ._utils import *


class PlaylistsMixin(MixinProtocol):
    def get_playlist(
        self, playlistId: str, limit: int | None = 100, related: bool = False, suggestions_limit: int = 0
    ) -> JsonDict:
        """
        Returns a list of playlist items

        :param playlistId: Playlist id
        :param limit: How many songs to return. ``None`` retrieves them all. Default: 100
        :param related: Whether to fetch 10 related playlists or not. Default: False
        :param suggestions_limit: How many suggestions to return. The result is a list of
            suggested playlist items (videos) contained in a "suggestions" key.
            7 items are retrieved in each internal request. Default: 0
        :return: Dictionary with information about the playlist.
            The key ``tracks`` contains a List of playlistItem dictionaries

        The result is in the following format::

            {
              "id": "PLQwVIlKxHM6qv-o99iX9R85og7IzF9YS_",
              "privacy": "PUBLIC",
              "title": "New EDM This Week 03/13/2020",
              "thumbnails": [...]
              "description": "Weekly r/EDM new release roundup. Created with github.com/sigma67/spotifyplaylist_to_gmusic",
              "author": {
                  "name": "sigmatics",
                  "id": "..."
              },
              "year": "2020",
              "duration": "6+ hours",
              "duration_seconds": 52651,
              "trackCount": 237,
              "suggestions": [
                  {
                    "videoId": "HLCsfOykA94",
                    "title": "Mambo (GATTÜSO Remix)",
                    "artists": [{
                        "name": "Nikki Vianna",
                        "id": "UCMW5eSIO1moVlIBLQzq4PnQ"
                      }],
                    "album": {
                      "name": "Mambo (GATTÜSO Remix)",
                      "id": "MPREb_jLeQJsd7U9w"
                    },
                    "likeStatus": "LIKE",
                    "thumbnails": [...],
                    "isAvailable": true,
                    "isExplicit": false,
                    "duration": "3:32",
                    "duration_seconds": 212,
                    "setVideoId": "to_be_updated_by_client"
                  }
              ],
              "related": [
                  {
                    "title": "Presenting MYRNE",
                    "playlistId": "RDCLAK5uy_mbdO3_xdD4NtU1rWI0OmvRSRZ8NH4uJCM",
                    "thumbnails": [...],
                    "description": "Playlist • YouTube Music"
                  }
              ],
              "tracks": [
                {
                  "videoId": "bjGppZKiuFE",
                  "title": "Lost",
                  "artists": [
                    {
                      "name": "Guest Who",
                      "id": "UCkgCRdnnqWnUeIH7EIc3dBg"
                    },
                    {
                      "name": "Kate Wild",
                      "id": "UCwR2l3JfJbvB6aq0RnnJfWg"
                    }
                  ],
                  "album": {
                    "name": "Lost",
                    "id": "MPREb_PxmzvDuqOnC"
                  },
                  "duration": "2:58",
                  "duration_seconds": 178,
                  "setVideoId": "748EE8..."
                  "likeStatus": "INDIFFERENT",
                  "thumbnails": [...],
                  "isAvailable": True,
                  "isExplicit": False,
                  "videoType": "MUSIC_VIDEO_TYPE_OMV",
                  "inLibrary": False,
                  "feedbackTokens": {
                    "add": "AB9zfpJxtvrU...",
                    "remove": "AB9zfpKTyZ..."
                  },
                  "pinnedToListenAgain": False,
                  "listenAgainFeedbackTokens": {
                    "pin": "AB9zfpImL2k...",
                    "unpin": "AB9zfpJt6pA..."
                  },
                  "communityVoteStatus": {
                    "netVoteValue": 12,
                    "status": "VOTE_STATUS_UPVOTED""
                  }
                }
                  "creditsBrowseId": "MPTCekz1IJ9I0sw"
              ]
            }

        The setVideoId is the unique id of this playlist item and
        needed for moving/removing playlist items

        Note that communityVoteStatus can be null if the information is not available (i.e. playlist does not have that setting enabled, the request is not authenticated, ...)

        Collaborative playlists replace ``author`` with limited data about ``collaborators``::
            {
                "collaborators": {
                    "text": "by Sample Author and 1 other",
                    "avatars": [
                        {
                            "url": "https://yt3.ggpht.com/sample-author-photo"
                        },
                        {
                            "url": "https://yt3.ggpht.com/sample-collaborator-photo"
                        }
                    ]
                }
            }
        """
        browseId = "VL" + playlistId if not playlistId.startswith("VL") else playlistId
        body = {"browseId": browseId}
        endpoint = "browse"
        request_func: RequestFuncType = lambda additionalParams: self._send_request(
            endpoint, body, additionalParams
        )
        response = request_func("")

        request_func_continuations: RequestFuncBodyType = lambda body: self._send_request(endpoint, body)
        is_ola = playlistId.startswith(("OLA", "VLOLA"))
        has_playlist_header = nav(response, [*TWO_COLUMN_RENDERER, *TAB_CONTENT, *SECTION_LIST_ITEM], True)
        if is_ola and not has_playlist_header:
            return parse_audio_playlist(response, limit, request_func_continuations)

        header_data = nav(response, [*TWO_COLUMN_RENDERER, *TAB_CONTENT, *SECTION_LIST_ITEM])
        section_list = nav(response, [*TWO_COLUMN_RENDERER, "secondaryContents", *SECTION])
        playlist: JsonDict = {}
        playlist["owned"] = EDITABLE_PLAYLIST_DETAIL_HEADER[0] in header_data
        if not playlist["owned"]:
            header = nav(header_data, RESPONSIVE_HEADER)
            playlist["id"] = nav(
                header,
                ["buttons", 1, "musicPlayButtonRenderer", "playNavigationEndpoint", *WATCH_PLAYLIST_ID],
                True,
            )
            playlist["privacy"] = "PUBLIC"
        else:
            playlist["id"] = nav(header_data, [*EDITABLE_PLAYLIST_DETAIL_HEADER, *PLAYLIST_ID])
            header = nav(header_data, [*EDITABLE_PLAYLIST_DETAIL_HEADER, *HEADER, *RESPONSIVE_HEADER])
            playlist["privacy"] = header_data[EDITABLE_PLAYLIST_DETAIL_HEADER[0]]["editHeader"][
                "musicPlaylistEditHeaderRenderer"
            ]["privacy"]

        description_shelf = nav(header, ["description", *DESCRIPTION_SHELF], True)
        playlist["description"] = (
            "".join([run["text"] for run in description_shelf["description"]["runs"]])
            if description_shelf
            else None
        )

        playlist.update(parse_playlist_header_meta(header))
        is_collaborative = "collaborators" in playlist

        playlist.update(parse_song_runs(nav(header, SUBTITLE_RUNS)[2 + playlist["owned"] * 2 :]))

        # suggestions and related are missing e.g. on liked songs
        playlist["related"] = []
        if "continuations" in section_list:
            additionalParams = get_continuation_params(section_list)
            if playlist["owned"] and (suggestions_limit > 0 or related):
                parse_func: ParseFuncType = lambda results: parse_playlist_items(results)
                suggested = request_func(additionalParams)
                continuation = nav(suggested, SECTION_LIST_CONTINUATION)
                additionalParams = get_continuation_params(continuation)
                suggestions_shelf = nav(continuation, CONTENT + MUSIC_SHELF)
                playlist["suggestions"] = get_continuation_contents(suggestions_shelf, parse_func)

                parse_func = lambda results: parse_playlist_items(results)
                playlist["suggestions"].extend(
                    get_reloadable_continuations(
                        suggestions_shelf,
                        "musicShelfContinuation",
                        suggestions_limit - len(playlist["suggestions"]),
                        request_func,
                        parse_func,
                    )
                )

            if related:
                response = request_func(additionalParams)
                continuation = nav(response, SECTION_LIST_CONTINUATION, True)
                if continuation:
                    parse_func = lambda results: parse_content_list(results, parse_playlist)
                    playlist["related"] = get_continuation_contents(
                        nav(continuation, CONTENT + CAROUSEL), parse_func
                    )

        playlist["tracks"] = []
        content_data = nav(section_list, [*CONTENT, "musicPlaylistShelfRenderer"])
        if "contents" in content_data:
            playlist["tracks"] = parse_playlist_items(
                content_data["contents"], is_collaborative=is_collaborative
            )

            parse_func = lambda contents: parse_playlist_items(contents, is_collaborative=is_collaborative)
            playlist["tracks"].extend(
                get_continuations_2025(content_data, limit, request_func_continuations, parse_func)
            )

        playlist["duration_seconds"] = sum_total_duration(playlist)
        return playlist

    def get_liked_songs(self, limit: int = 100) -> JsonDict:
        """
        Gets playlist items for the 'Liked Songs' playlist

        :param limit: How many items to return. Default: 100
        :return: List of playlistItem dictionaries. See :py:func:`get_playlist`
        """
        self._check_auth()
        return self.get_playlist("LM", limit)

    def get_saved_episodes(self, limit: int = 100) -> JsonDict:
        """
        Gets playlist items for the 'Liked Songs' playlist

        :param limit: How many items to return. Default: 100
        :return: List of playlistItem dictionaries. See :py:func:`get_playlist`
        """
        return self.get_playlist("SE", limit)

    def create_playlist(
        self,
        title: str,
        description: str,
        privacy_status: str = "PRIVATE",
        video_ids: list[str] | None = None,
        source_playlist: str | None = None,
    ) -> str | JsonDict:
        """
        Creates a new empty playlist and returns its id.

        :param title: Playlist title
        :param description: Playlist description
        :param privacy_status: Playlists can be ``PUBLIC``, ``PRIVATE``, or ``UNLISTED``. Default: ``PRIVATE``
        :param video_ids: IDs of songs to create the playlist with
        :param source_playlist: Another playlist whose songs should be added to the new playlist
        :return: ID of the YouTube playlist or full response if there was an error
        """
        self._check_auth()

        invalid_characters = ["<", ">"]  # ytmusic will crash if these are part of the title
        invalid_characters_found = [invalid for invalid in invalid_characters if invalid in title]
        if invalid_characters_found:
            msg = f"{title} contains invalid characters: {', '.join(invalid_characters_found)}"
            raise YTMusicUserError(msg)

        body: JsonDict = {
            "title": title,
            "description": html_to_txt(description),  # YT does not allow HTML tags
            "privacyStatus": privacy_status,
        }
        if video_ids is not None:
            body["videoIds"] = video_ids

        if source_playlist is not None:
            body["sourcePlaylistId"] = source_playlist

        endpoint = "playlist/create"
        response = self._send_request(endpoint, body)
        if "playlistId" in response:
            playlist_id: str = response["playlistId"]
            return playlist_id

        validate_write_response(response)
        return response

    def join_collaborative_playlist(self, playlistId: str, joinCollaborationToken: str) -> str | JsonDict:
        """
        Given an invite token, join a collaborative playlist and add it to your library.

        :param playlistId: ID of the playlist to join.
            If you're already a collaborator, an Unauthorized server error is raised.
        :param joinCollaborationToken: See :py:func:`edit_playlist`, or ``jct`` in YTM's invite URL
        :return: Status String or full response
        """
        self._check_auth()
        body: JsonDict = {
            "playlistId": validate_playlist_id(playlistId),
            "actions": [
                {"action": "ACTION_JOIN_COLLABORATION", "joinCollaborationToken": joinCollaborationToken}
            ],
        }
        endpoint = "browse/edit_playlist"
        response = self._send_request(endpoint, body)
        result: str | JsonDict = response.get("status", response)
        return result

    def edit_playlist(
        self,
        playlistId: str,
        title: str | None = None,
        description: str | None = None,
        privacyStatus: str | None = None,
        collaboration: bool | None = None,
        moveItem: str | tuple[str, str] | None = None,
        addPlaylistId: str | None = None,
        sortOrder: PlaylistSortOrder | None = None,
        addToTop: bool | None = None,
        voteOption: PlaylistVoteEditOptions | None = None,
    ) -> str | JsonDict:
        """
        Edit title, description or privacyStatus of a playlist.
        You may also move an item within a playlist or append another playlist to this playlist.

        :param playlistId: Playlist id
        :param title: Optional. New title for the playlist
        :param description: Optional. New description for the playlist
        :param privacyStatus: Optional. New privacy status for the playlist
        :param collaboration: Optional. Enable or disable collaboration.
            If False and collaboration is not enabled, a Forbidden server error is raised.
            If True, a new ``joinCollaborationToken`` is returned.
            Collaborators cannot interact with private playlists.
        :param moveItem: Optional. Move one item before another. Items are specified by setVideoId, which is the
            unique id of this playlist item. See :py:func:`get_playlist`
        :param addPlaylistId: Optional. Id of another playlist to add to this playlist
        :param sortOrder: Optional. Change the order tracks are returned in. The default is ``MANUAL``.
        :param addToTop: Optional. Change the state of this playlist to add items to the top of the playlist (if True)
            or the bottom of the playlist (if False - this is also the default of a new playlist).
        :param VoteOption: Optional. Change who can participate in community voting in this playlist.
            Note that a bad request will be thrown if voteOption is PlaylistVoteEditOptions.COLLABORATORS_ONLY but the playlist is not enabled for collaboration prior to the edit.
        :return: Status String, ``collaboration`` dict described below, or full response

        Dictionary returned when ``collaboration`` is True and the request is successful::

                {
                    "status": "STATUS_SUCCEEDED",
                    "joinCollaborationToken": "kM9wXdRj2p8v_qL3sHBkTz"
                }
        """
        self._check_auth()
        body: JsonDict = {"playlistId": validate_playlist_id(playlistId)}
        actions: JsonList = []
        if title:
            actions.append({"action": "ACTION_SET_PLAYLIST_NAME", "playlistName": title})

        if description:
            actions.append({"action": "ACTION_SET_PLAYLIST_DESCRIPTION", "playlistDescription": description})

        if privacyStatus:
            actions.append({"action": "ACTION_SET_PLAYLIST_PRIVACY", "playlistPrivacy": privacyStatus})

        if collaboration:
            actions.append({"action": "ACTION_CREATE_COLLABORATION_INVITE_LINK"})
        elif collaboration is False:
            actions.append({"action": "ACTION_SET_CLOSED_TO_CONTRIBUTIONS", "closedToContributions": True})

        if moveItem:
            action = {
                "action": "ACTION_MOVE_VIDEO_BEFORE",
                "setVideoId": moveItem if isinstance(moveItem, str) else moveItem[0],
            }
            if isinstance(moveItem, tuple) and len(moveItem) > 1:
                action["movedSetVideoIdSuccessor"] = moveItem[1]
            actions.append(action)

        if addPlaylistId:
            actions.append({"action": "ACTION_ADD_PLAYLIST", "addedFullListId": addPlaylistId})

        if sortOrder:
            actions.append(
                {"action": "ACTION_SET_PLAYLIST_VIDEO_ORDER", "playlistVideoOrder": sortOrder.value}
            )

        if addToTop is not None:
            actions.append({"action": "ACTION_SET_ADD_TO_TOP", "addToTop": str(addToTop).lower()})

        if voteOption is not None:
            actions.append(
                {
                    "action": "ACTION_SET_ALLOW_ITEM_VOTE",
                    "itemVotePermission": voteOption.get_argument_for_request(),
                }
            )

        body["actions"] = actions
        endpoint = "browse/edit_playlist"
        response = self._send_request(endpoint, body)

        if collaboration and response.get("status") == ResponseStatus.SUCCEEDED:
            invite_link = nav(response, ["collaborationInviteLink"])
            return {
                "status": response["status"],
                "joinCollaborationToken": nav(parse_qs(urlparse(invite_link).query), ["jct", 0]),
            }

        result: str | JsonDict = response.get("status", response)
        return result

    def delete_playlist(self, playlistId: str) -> str | JsonDict:
        """
        Delete a playlist.

        :param playlistId: Playlist id
        :return: Status String or full response
        """
        self._check_auth()
        body = {"playlistId": validate_playlist_id(playlistId)}
        endpoint = "playlist/delete"
        response = self._send_request(endpoint, body)
        result: str | JsonDict = response.get("status", response)
        return result

    def add_playlist_items(
        self,
        playlistId: str,
        videoIds: list[str] | None = None,
        source_playlist: str | None = None,
        duplicates: bool = False,
    ) -> str | JsonDict:
        """
        Add songs to an existing playlist

        :param playlistId: Playlist id
        :param videoIds: List of Video ids
        :param source_playlist: Playlist id of a playlist to add to the current playlist (no duplicate check)
        :param duplicates: If True, duplicates will be added. If False, an error will be returned if there are duplicates (no items are added to the playlist)
        :return: Status String and a dict containing the new setVideoId for each videoId or full response
        """
        self._check_auth()
        body: JsonDict = {"playlistId": validate_playlist_id(playlistId), "actions": []}
        if not videoIds and not source_playlist:
            raise YTMusicUserError(
                "You must provide either videoIds or a source_playlist to add to the playlist"
            )

        if videoIds:
            for videoId in videoIds:
                action = {"action": "ACTION_ADD_VIDEO", "addedVideoId": videoId}
                if duplicates:
                    action["dedupeOption"] = "DEDUPE_OPTION_SKIP"
                body["actions"].append(action)

        if source_playlist:
            body["actions"].append({"action": "ACTION_ADD_PLAYLIST", "addedFullListId": source_playlist})

            # add an empty ACTION_ADD_VIDEO because otherwise
            # YTM doesn't return the dict that maps videoIds to their new setVideoIds
            if not videoIds:
                body["actions"].append({"action": "ACTION_ADD_VIDEO", "addedVideoId": None})

        endpoint = "browse/edit_playlist"
        response = self._send_request(endpoint, body)
        if "status" in response and "SUCCEEDED" in response["status"]:
            result_dict = [
                result_data.get("playlistEditVideoAddedResultData")
                for result_data in response.get("playlistEditResults", [])
            ]
            return {"status": response["status"], "playlistEditResults": result_dict}
        else:
            return response

    def remove_playlist_items(self, playlistId: str, videos: JsonList) -> str | JsonDict:
        """
        Remove songs from an existing playlist

        :param playlistId: Playlist id
        :param videos: List of PlaylistItems, see :py:func:`get_playlist`.
            Must contain videoId and setVideoId
        :return: Status String or full response
        """
        self._check_auth()
        videos = list(filter(lambda x: "videoId" in x and "setVideoId" in x, videos))
        if len(videos) == 0:
            raise YTMusicUserError(
                "Cannot remove songs, because setVideoId is missing. Do you own this playlist?"
            )

        body: JsonDict = {"playlistId": validate_playlist_id(playlistId), "actions": []}
        for video in videos:
            body["actions"].append(
                {
                    "setVideoId": video["setVideoId"],
                    "removedVideoId": video["videoId"],
                    "action": "ACTION_REMOVE_VIDEO",
                }
            )

        endpoint = "browse/edit_playlist"
        response = self._send_request(endpoint, body)
        result: str | JsonDict = response.get("status", response)
        return result



from ytmusicapi.mixins._protocol import MixinProtocol
from ytmusicapi.parsers.explore import *
from ytmusicapi.type_alias import JsonDict, JsonList

# extra playlist carousel some countries return between the video charts and the artists
COUNTRY_EXTRA_CATEGORY = {"US": "genres", "IN": "languages"}


def is_playlist_carousel(contents: JsonList) -> bool:
    """chart playlists are the only carousel items linking to a "VL" playlist"""
    browse_id = nav(contents[0], [MTRIR, *TITLE, *NAVIGATION_BROWSE_ID], True)
    return bool(browse_id and browse_id.startswith("VL"))


def is_artist_carousel(contents: JsonList) -> bool:
    return MRLIR in contents[0]


class ChartsMixin(MixinProtocol):
    def get_charts(self, country: str = "ZZ") -> JsonDict:
        """
        Get latest charts data from YouTube Music: Artists and playlists of top videos.
        Unauthenticated requests return unranked Artists with "rank" and "trend" set to None.
        US charts have an extra Genres section, IN charts an extra Languages section.

        :param country: ISO 3166-1 Alpha-2 country code. Default: ``ZZ`` = Global
        :return: Dictionary containing chart video playlists (with separate daily/weekly charts if authenticated with a premium account),
            chart genres (US-only, not returned by every response), chart languages (IN-only) and chart artists.

        Example::

            {
                "countries": {
                    "selected": {
                        "text": "United States"
                    },
                    "options": ["DE",
                        "ZZ",
                        "ZW"]
                },
                "videos": [
                    {
                        "title": "Daily Top Music Videos - United States",
                        "playlistId": "PL4fGSI1pDJn61unMfmrUSz68RT8IFFnks",
                        "thumbnails": []
                    }
                ],
                "artists": [
                    {
                        "title": "YoungBoy Never Broke Again",
                        "browseId": "UCR28YDxjDE3ogQROaNdnRbQ",
                        "subscribers": "9.62M",
                        "thumbnails": [],
                        "rank": "1",
                        "trend": "neutral"
                    }
                ],
                "genres": [
                    {
                        "title": "Top 50 Pop Music Videos United States",
                        "playlistId": "PL4fGSI1pDJn77aK7sAW2AT0oOzo5inWY8",
                        "thumbnails": []
                    }
                ]
            }

        """
        body: JsonDict = {"browseId": "FEmusic_charts"}
        if country:
            body["formData"] = {"selectedValues": [country]}

        response = self._send_request("browse", body)
        results = nav(response, SINGLE_COLUMN_TAB + SECTION_LIST)
        charts: JsonDict = {"countries": {}}
        menu = nav(
            results[0],
            [
                *MUSIC_SHELF,
                "subheaders",
                0,
                "musicSideAlignedItemRenderer",
                "startItems",
                0,
                "musicSortFilterButtonRenderer",
            ],
        )
        charts["countries"]["selected"] = nav(menu, TITLE)
        charts["countries"]["options"] = list(
            filter(
                None,
                [
                    nav(m, ["payload", "musicFormBooleanChoice", "opaqueToken"], True)
                    for m in nav(response, FRAMEWORK_MUTATIONS)
                ],
            )
        )

        # carousels carry no machine-readable identity, so recognize them by their contents;
        # anything else (some regions add an album chart) is ignored instead of shifting the
        # categories that come after it
        carousels = [c for section in results[1:] if (c := nav(section, CAROUSEL_CONTENTS, True))]
        playlist_carousels = [c for c in carousels if is_playlist_carousel(c)]
        artist_carousels = [c for c in carousels if is_artist_carousel(c)]

        playlist_names = ["videos"]
        if country in COUNTRY_EXTRA_CATEGORY:
            playlist_names.append(COUNTRY_EXTRA_CATEGORY[country])
        # premium sessions get daily and weekly carousels in place of the "videos" one
        # could also be done via an is_premium attribute on YTMusic instance
        if len(playlist_carousels) > len(playlist_names):
            playlist_names = ["daily", "weekly", *playlist_names[1:]]

        # zip leaves trailing names unused: YTM omits categories unpredictably,
        # e.g. "genres" is missing from some US responses
        for name, contents in zip(playlist_names, playlist_carousels):
            charts[name] = parse_content_list(contents, parse_chart_playlist, MTRIR)

        if artist_carousels:
            charts["artists"] = parse_content_list(artist_carousels[0], parse_chart_artist, MRLIR)

        return charts



import re

from ytmusicapi.type_alias import JsonDict, JsonList

from ._utils import *
from .artists import parse_artists_runs
from .constants import DOT_SEPARATOR_RUN


def parse_song_artists(data: JsonDict, index: int) -> JsonList:
    flex_item = get_flex_column_item(data, index)
    if not flex_item:
        return []
    else:
        runs = flex_item["text"]["runs"]
        return parse_artists_runs(runs)


# leading views word ("조회수 17억회", "播放次數：4505") or bidi mark (ur), up to the first digit  # noqa: RUF003
VIEWS_PREFIX = re.compile(r"^\D*?[\s:\uff1a\u200e-\u200f\u202a-\u202e]")


def parse_views(text: str) -> str | None:
    # only for non-latin scripts: "Maroon 5" is indistinguishable from a prefixed count
    prefixed = 0
    if not re.search(r"[a-zA-Z]", text):
        text, prefixed = VIEWS_PREFIX.subn("", text)

    if not re.match(r"^\d", text):
        return None

    # a bare ASCII token like "2Pac" is an artist, not a view count with a stripped word
    if not prefixed and text.isascii() and " " not in text:
        return None

    # note: \xa0 glues number and magnitude ("1,7\xa0Mrd. Aufrufe"), but some locales use
    # it before the views word too ("88\xa0k\xa0vues"); CJK has no separator ("3406万回視聴")
    head = text.split(" ")[0].split("\xa0")
    return "\xa0".join(head[:2])


def parse_song_run(run: JsonDict) -> JsonDict:
    text = run["text"]

    if "navigationEndpoint" in run:  # artist or album
        item = {"name": text, "id": nav(run, NAVIGATION_BROWSE_ID, True)}

        if item["id"] and (item["id"].startswith("MPRE") or "release_detail" in item["id"]):  # album
            return {"type": "album", "data": item}
        else:  # artist
            return {"type": "artist", "data": item}
    else:
        if re.match(r"^(\d+:)*\d+:\d+$", text):
            return {"type": "duration", "data": text}

        elif re.match(r"^\d{4}$", text):
            return {"type": "year", "data": text}

        elif (views := parse_views(text)) is not None:
            return {"type": "views", "data": views}

        else:  # artist without id
            return {"type": "artist", "data": {"name": text, "id": None}}


def parse_song_runs(runs: JsonList, skip_type_spec: bool = False) -> JsonDict:
    """
    :param skip_type_spec: if true, skip the type specifier (like "Song", "Single", or "Album") that may appear before artists ("Song • Eminem"). Otherwise, that text item is parsed as an artist with no ID.
    """

    parsed: JsonDict = {}

    # prevent type specifier from being parsed as an artist
    # it's the unlinked first run, separated by " • " from the artists, or from
    # metadata (duration/views/year) when the song lists no artist at all
    if (
        skip_type_spec
        and len(runs) > 2
        and "navigationEndpoint" not in runs[0]
        and parse_song_run(runs[0])["type"] == "artist"
        and runs[1] == DOT_SEPARATOR_RUN
        and parse_song_run(runs[2])["type"] in ("artist", "duration", "views", "year")
    ):
        runs = runs[2:]

    for i, run in list(enumerate(runs)):
        if i % 2:  # uneven items are always separators
            continue

        parsed_run = parse_song_run(run)
        data = parsed_run["data"]
        match parsed_run["type"]:
            case "album":
                parsed["album"] = data
            case "artist":
                parsed["artists"] = parsed.get("artists", [])
                parsed["artists"].append(data)
            case "views":
                parsed["views"] = data
            case "duration":
                parsed["duration"] = data
                parsed["duration_seconds"] = parse_duration(data)
            case "year":
                parsed["year"] = data

    return parsed


def parse_song_album(data: JsonDict, index: int) -> JsonDict | None:
    flex_item = get_flex_column_item(data, index)
    browse_id = nav(flex_item, TEXT_RUN + NAVIGATION_BROWSE_ID, True)
    return None if not flex_item else {"name": get_item_text(data, index), "id": browse_id}


def parse_song_menu_data(data: JsonDict) -> JsonDict:
    """
    :return: Dictionary with data from the provided song's context menu.

    Example::

        {
            "inLibrary": true,
            "feedbackTokens": {
                "add": "...",
                "remove": "..."
            },
            "pinnedToListenAgain": true,
            "listenAgainFeedbackTokens": {
                "pin": "...",
                "unpin": "..."
            }
        }
    """

    if "menu" not in data:
        return {}

    song_data: JsonDict = {}
    for item in nav(data, MENU_ITEMS):
        menu_item = nav(item, [TOGGLE_MENU], True) or nav(item, ["menuServiceItemRenderer"], True)
        if menu_item is None:
            continue

        song_data["inLibrary"] = song_data.get("inLibrary", False)
        song_data["pinnedToListenAgain"] = song_data.get("pinnedToListenAgain", False)

        current_icon_type = nav(menu_item, ["defaultIcon", "iconType"], True) or nav(
            menu_item, ["icon", "iconType"], True
        )

        def feedback_token(endpoint_type: str, menu_item: JsonDict = menu_item) -> str | None:
            return nav(menu_item, [endpoint_type, *FEEDBACK_TOKEN], True)

        # YTM signals the current state with isToggled instead of swapping the default/toggled icons
        is_toggled = bool(menu_item.get("isToggled"))

        match current_icon_type:
            case "KEEP":  # pin to listen again
                song_data["pinnedToListenAgain"] = is_toggled
                song_data["listenAgainFeedbackTokens"] = {
                    "pin": feedback_token("defaultServiceEndpoint"),
                    "unpin": feedback_token("toggledServiceEndpoint"),
                }
            case "KEEP_OFF":  # unpin from listen again
                song_data["pinnedToListenAgain"] = True
                song_data["listenAgainFeedbackTokens"] = {
                    "pin": feedback_token("toggledServiceEndpoint"),
                    "unpin": feedback_token("defaultServiceEndpoint"),
                }
            case "BOOKMARK_BORDER":  # add to library
                song_data["inLibrary"] = is_toggled
                song_data["feedbackTokens"] = {
                    "add": feedback_token("defaultServiceEndpoint"),
                    "remove": feedback_token("toggledServiceEndpoint"),
                }
            case "BOOKMARK":  # remove from library
                song_data["inLibrary"] = True
                song_data["feedbackTokens"] = {
                    "add": feedback_token("toggledServiceEndpoint"),
                    "remove": feedback_token("defaultServiceEndpoint"),
                }
            case "REMOVE_FROM_HISTORY":
                song_data["feedbackToken"] = feedback_token("serviceEndpoint")

    return song_data


def parse_like_status(service: JsonDict) -> str:
    status = ["LIKE", "INDIFFERENT"]
    return status[status.index(service["likeEndpoint"]["status"]) - 1]



from enum import Enum
from typing import Any


class PrivacyStatus(str, Enum):
    PUBLIC = "PUBLIC"
    PRIVATE = "PRIVATE"
    UNLISTED = "UNLISTED"


class LikeStatus(str, Enum):
    LIKE = "LIKE"
    DISLIKE = "DISLIKE"
    INDIFFERENT = "INDIFFERENT"

    @classmethod
    def _missing_(cls, value: Any) -> "LikeStatus":
        return cls.INDIFFERENT


class VideoType(str, Enum):
    OMV = "MUSIC_VIDEO_TYPE_OMV"
    UGC = "MUSIC_VIDEO_TYPE_UGC"
    ATV = "MUSIC_VIDEO_TYPE_ATV"
    OFFICIAL_SOURCE_MUSIC = "MUSIC_VIDEO_TYPE_OFFICIAL_SOURCE_MUSIC"


class PlaylistSortOrder(Enum):
    MANUAL = 0
    NEWEST_FIRST = 1
    NEWEST_LAST = 2
    TOP_VOTED = 6


class VoteStatus(str, Enum):
    UPVOTED = "VOTE_STATUS_UPVOTED"
    DOWNVOTED = "VOTE_STATUS_DOWNVOTED"
    UNSPECIFIED = "VOTE_STATUS_UNSPECIFIED"


class PlaylistVoteEditOptions(str, Enum):
    EVERYONE_CAN_VOTE = "EVERYONE_CAN_VOTE"
    # only available for playlists where collaborate is on.
    COLLABORATORS_ONLY = "COLLABORATORS_ONLY"
    OFF = "OFF"

    def get_argument_for_request(self) -> int:
        match self:
            case PlaylistVoteEditOptions.EVERYONE_CAN_VOTE:
                return 1
            case PlaylistVoteEditOptions.COLLABORATORS_ONLY:
                return 2
            case PlaylistVoteEditOptions.OFF:
                return 3



from ytmusicapi.type_alias import JsonDict, JsonList

from .songs import *


def parse_watch_playlist(results: JsonList) -> JsonList:
    tracks = []
    PPVWR = "playlistPanelVideoWrapperRenderer"
    PPVR = "playlistPanelVideoRenderer"
    for result in results:
        counterpart = None
        if PPVWR in result:
            counterpart = result[PPVWR]["counterpart"][0]["counterpartRenderer"][PPVR]
            result = result[PPVWR]["primaryRenderer"]
        if PPVR not in result:
            continue
        data = result[PPVR]
        if "unplayableText" in data:
            continue

        track = parse_watch_track(data)
        if counterpart:
            track["counterpart"] = parse_watch_track(counterpart)
        tracks.append(track)

    return tracks


def parse_watch_track(data: JsonDict) -> JsonDict:
    like_status = None
    for item in nav(data, MENU_ITEMS):
        if TOGGLE_MENU in item:
            service = item[TOGGLE_MENU]["defaultServiceEndpoint"]
            if "likeEndpoint" in service:
                like_status = parse_like_status(service)

    track = {
        "videoId": data["videoId"],
        "title": nav(data, TITLE_TEXT),
        "length": nav(data, ["lengthText", "runs", 0, "text"], True),
        "thumbnail": nav(data, THUMBNAIL),
        "likeStatus": like_status,
        "videoType": nav(data, ["navigationEndpoint", *NAVIGATION_VIDEO_TYPE], True),
    }

    track.update(
        {
            "inLibrary": None,
            "feedbackTokens": None,
            "pinnedToListenAgain": None,
            "listenAgainFeedbackTokens": None,
        }
        | parse_song_menu_data(data)
    )

    if longBylineText := nav(data, ["longBylineText"]):
        song_info = parse_song_runs(longBylineText["runs"])
        track.update(song_info)

    return track


def get_tab_browse_ids(watchNextRenderer: JsonDict) -> dict[str, str]:
    browse_ids = {}
    for tab in watchNextRenderer["tabs"]:
        if "unselectable" in tab["tabRenderer"]:
            continue

        if browse_endpoint := nav(tab, ["tabRenderer", "endpoint", "browseEndpoint"], True):
            page_type = nav(browse_endpoint, PAGE_TYPE)
            browse_ids[page_type] = browse_endpoint["browseId"]
    return browse_ids



import re
from datetime import datetime, timezone
from typing import Literal

from ytmusicapi.exceptions import YTMusicGatedError, YTMusicUserError
from ytmusicapi.models.content.enums import LikeStatus
from ytmusicapi.navigation import nav
from ytmusicapi.type_alias import JsonDict

LibraryOrderType = Literal["a_to_z", "z_to_a", "recently_added"]


def prepare_like_endpoint(rating: str | LikeStatus) -> str:
    match rating:
        case LikeStatus.LIKE:
            return "like/like"
        case LikeStatus.DISLIKE:
            return "like/dislike"
        case LikeStatus.INDIFFERENT:
            return "like/removelike"
        case _:
            raise YTMusicUserError(
                f"Invalid rating provided. Use one of {[e.value for e in LikeStatus.__members__.values()]}."
            )


def validate_order_parameter(order: LibraryOrderType | None) -> None:
    """Validate the provided order, if any

    :raises YTMusicUserError: if the provided order is invalid
    """
    orders = ["a_to_z", "z_to_a", "recently_added"]
    if order and order not in orders:
        raise YTMusicUserError(
            "Invalid order provided. Please use one of the following orders or leave out the parameter: "
            + ", ".join(orders)
        )


def prepare_order_params(order: LibraryOrderType) -> str:
    """Returns request params belonging to a specific sorting order."""
    orders = ["a_to_z", "z_to_a", "recently_added"]
    # determine order_params via `.contents.singleColumnBrowseResultsRenderer.tabs[0].tabRenderer.content.sectionListRenderer.contents[1].itemSectionRenderer.header.itemSectionTabbedHeaderRenderer.endItems[1].dropdownRenderer.entries[].dropdownItemRenderer.onSelectCommand.browseEndpoint.params` of `/youtubei/v1/browse` response
    order_params = ["ggMGKgQIARAA", "ggMGKgQIARAB", "ggMGKgQIABAB"]
    return order_params[orders.index(order)]


def html_to_txt(html_text: str) -> str:
    """
    Sanitize tags from html

    :param html_text: String containing html tags.
    :return: String without < > characters
    """
    tags = re.findall("<[^>]+>", html_text)
    for tag in tags:
        html_text = html_text.replace(tag, "")
    return html_text


def validate_write_response(response: JsonDict) -> None:
    """Check that YTM performed a write request instead of answering with a dialog.

    A gated request returns HTTP 200 with only a ``showEngagementPanelEndpoint`` action carrying the
    original request, for the web client to replay once the user has dealt with the dialog.

    :raises YTMusicGatedError: if the request was not performed
    """
    panel = nav(response, ["actions", 0, "showEngagementPanelEndpoint"], True)
    if panel is None:
        return

    tag = nav(panel, ["identifier", "tag"], True)
    raise YTMusicGatedError(
        f"YouTube Music did not perform this request and asked for interaction with its '{tag}' dialog. "
        "The account may be temporarily restricted from this action."
    )


def get_datestamp() -> int:
    """Returns the number of days since January 1, 1970.
    Currently only used for the signature timestamp in :py:func:`get_song`."""
    epoch = datetime(1970, 1, 1, tzinfo=timezone.utc)
    return (datetime.now(timezone.utc) - epoch).days



import re

from ytmusicapi.continuations import *
from ytmusicapi.helpers import sum_total_duration
from ytmusicapi.models.content.enums import VoteStatus
from ytmusicapi.type_alias import JsonDict, JsonList, ParseFuncType, RequestFuncBodyType

from ..helpers import to_int
from .songs import *


def parse_playlist_header(response: JsonDict) -> JsonDict:
    playlist: JsonDict = {}
    editable_header = nav(response, [*HEADER, *EDITABLE_PLAYLIST_DETAIL_HEADER], True)
    playlist["owned"] = editable_header is not None
    playlist["privacy"] = "PUBLIC"
    if editable_header is not None:  # owned playlist
        header = nav(editable_header, HEADER_DETAIL)
        playlist["privacy"] = editable_header["editHeader"]["musicPlaylistEditHeaderRenderer"]["privacy"]
    else:
        header = nav(response, HEADER_DETAIL, True)
        if header is None:
            header = nav(
                response, [*TWO_COLUMN_RENDERER, *TAB_CONTENT, *SECTION_LIST_ITEM, *RESPONSIVE_HEADER]
            )

    playlist.update(parse_playlist_header_meta(header))
    if playlist["thumbnails"] is None:
        playlist["thumbnails"] = nav(header, THUMBNAIL_CROPPED, True)
    playlist["description"] = nav(header, ["description", *DESCRIPTION_SHELF, *DESCRIPTION], True)
    playlist["year"] = nav(header, SUBTITLE2)

    return playlist


def parse_playlist_header_meta(header: JsonDict) -> JsonDict:
    playlist_meta = {
        "views": None,
        "duration": None,
        "trackCount": None,
        "title": "".join([run["text"] for run in header.get("title", {}).get("runs", [])]),
        "thumbnails": nav(header, THUMBNAILS),
    }
    if "facepile" in header:
        avatar_renderer = nav(header, ["facepile", "avatarStackViewModel", "rendererContext"])
        avatar_command = nav(
            avatar_renderer,
            [
                "commandContext",
                "onTap",
                "innertubeCommand",
            ],
            True,
        )

        if (
            nav(avatar_command, ["showEngagementPanelEndpoint", "identifier", "tag"], True)
            == "PAplaylist_collaborate"
        ):
            avatars = nav(header, ["facepile", "avatarStackViewModel", "avatars"])

            playlist_meta["collaborators"] = {
                "text": nav(avatar_renderer, ["accessibilityContext", "label"]),
                "avatars": [avatar["avatarViewModel"]["image"]["sources"][0] for avatar in avatars],
            }
        else:
            playlist_meta["author"] = {
                "name": nav(header, ["facepile", "avatarStackViewModel", "text", "content"]),
                "id": nav(
                    avatar_command,
                    [
                        "browseEndpoint",
                        "browseId",
                    ],
                    True,
                ),
            }
    if "runs" in header["secondSubtitle"]:
        second_subtitle_runs = header["secondSubtitle"]["runs"]
        has_views = (len(second_subtitle_runs) > 3) * 2
        playlist_meta["views"] = None if not has_views else to_int(second_subtitle_runs[0]["text"])
        has_duration = (len(second_subtitle_runs) > 1) * 2
        playlist_meta["duration"] = (
            None if not has_duration else second_subtitle_runs[has_views + has_duration]["text"]
        )
        song_count_text = second_subtitle_runs[has_views + 0]["text"]
        song_count_search = re.findall(r"\d+", song_count_text)
        # extract the digits from the text, return 0 if no match
        playlist_meta["trackCount"] = (
            to_int("".join(song_count_search)) if song_count_search is not None else None
        )

    return playlist_meta


def parse_audio_playlist(
    response: JsonDict, limit: int | None, request_func: RequestFuncBodyType
) -> JsonDict:
    playlist: JsonDict = {
        "owned": False,
        "privacy": "PUBLIC",
        "description": None,
        "views": None,
        "duration": None,
        "tracks": [],
        "thumbnails": [],
        "related": [],
    }

    section_list = nav(response, [*TWO_COLUMN_RENDERER, "secondaryContents", *SECTION])
    content_data = nav(section_list, [*CONTENT, "musicPlaylistShelfRenderer"])

    playlist["id"] = nav(content_data, ["targetId"])

    playlist["tracks"] = []
    if "contents" in content_data:
        playlist["tracks"] = parse_playlist_items(content_data["contents"])

        parse_func: ParseFuncType = lambda contents: parse_playlist_items(contents)
        playlist["tracks"].extend(get_continuations_2025(content_data, limit, request_func, parse_func))

    playlist["trackCount"] = len(playlist["tracks"])

    playlist["title"] = playlist["tracks"][0]["album"]["name"]

    playlist["duration_seconds"] = sum_total_duration(playlist)
    return playlist


def parse_playlist_items(
    results: JsonList,
    is_album: bool = False,
    is_collaborative: bool = False,
) -> JsonList:
    songs = []
    for result in results:
        if MRLIR not in result:
            continue
        data = result[MRLIR]
        song = parse_playlist_item(data, is_album, is_collaborative)
        if song:
            songs.append(song)

    return songs


def parse_playlist_item(
    data: JsonDict,
    is_album: bool = False,
    is_collaborative: bool = False,
) -> JsonDict | None:
    videoId = setVideoId = creditsBrowseId = None
    like = None

    # if the item has a menu, find its setVideoId
    if "menu" in data:
        for item in nav(data, MENU_ITEMS):
            if "menuServiceItemRenderer" in item:
                menu_service = nav(item, MENU_SERVICE)
                if "playlistEditEndpoint" in menu_service:
                    setVideoId = nav(menu_service, ["playlistEditEndpoint", "actions", 0, "setVideoId"], True)
                    videoId = nav(
                        menu_service, ["playlistEditEndpoint", "actions", 0, "removedVideoId"], True
                    )
            elif MNIR in item:
                maybe_credits_browse_id = nav(item, [MNIR, *NAVIGATION_BROWSE_ID], True)
                if maybe_credits_browse_id and maybe_credits_browse_id.startswith("MPTC"):
                    creditsBrowseId = maybe_credits_browse_id

    song_menu_data = {"inLibrary": None, "pinnedToListenAgain": None} | parse_song_menu_data(data)

    # if item is not playable, the videoId was retrieved above
    if nav(data, PLAY_BUTTON, none_if_absent=True) is not None and "playNavigationEndpoint" in nav(
        data, PLAY_BUTTON
    ):
        videoId = nav(data, PLAY_BUTTON)["playNavigationEndpoint"]["watchEndpoint"]["videoId"]

        if "menu" in data:
            like = nav(data, MENU_LIKE_STATUS, True)

    isAvailable = True
    if "musicItemRendererDisplayPolicy" in data:
        isAvailable = data["musicItemRendererDisplayPolicy"] != "MUSIC_ITEM_RENDERER_DISPLAY_POLICY_GREY_OUT"

    # For unavailable items and for album track lists indexes are preset,
    # because meaning of the flex column cannot be reliably found using navigationEndpoint
    use_preset_columns = True if isAvailable is False or is_album is True else None

    title_index = 0 if use_preset_columns else None
    artist_index = 1 if use_preset_columns else None
    duration_index = None
    # collaborative playlists have duration in flexColumns (between artist and album)
    album_index = 3 if is_collaborative else 2 if use_preset_columns else None
    user_channel_indexes = []
    unrecognized_index = None

    for index in range(len(data["flexColumns"])):
        flex_column_item = get_flex_column_item(data, index)
        navigation_endpoint = nav(flex_column_item, [*TEXT_RUN, "navigationEndpoint"], True)

        if not navigation_endpoint:
            run = nav(flex_column_item, TEXT_RUN, True)
            if run and "text" in run:
                parsed = parse_song_run(run)
                if parsed["type"] == "duration":
                    duration_index = index
                else:
                    unrecognized_index = index if unrecognized_index is None else unrecognized_index

            continue

        if "watchEndpoint" in navigation_endpoint:
            title_index = index
        elif "browseEndpoint" in navigation_endpoint:
            page_type = nav(
                navigation_endpoint,
                [
                    "browseEndpoint",
                    "browseEndpointContextSupportedConfigs",
                    "browseEndpointContextMusicConfig",
                    "pageType",
                ],
            )

            # MUSIC_PAGE_TYPE_ARTIST for regular songs, MUSIC_PAGE_TYPE_UNKNOWN for uploads
            if page_type == "MUSIC_PAGE_TYPE_ARTIST" or page_type == "MUSIC_PAGE_TYPE_UNKNOWN":
                artist_index = index
            elif page_type in ["MUSIC_PAGE_TYPE_ALBUM", "MUSIC_PAGE_TYPE_AUDIOBOOK"]:
                album_index = index
            elif page_type == "MUSIC_PAGE_TYPE_USER_CHANNEL":
                user_channel_indexes.append(index)
            # Non music videos, for example: podcast episodes
            elif page_type == "MUSIC_PAGE_TYPE_NON_MUSIC_AUDIO_TRACK_PAGE":
                title_index = index

    # Extra check for rare songs, where artist is non-clickable and does not have navigationEndpoint
    if artist_index is None and unrecognized_index is not None:
        artist_index = unrecognized_index

    # Extra check for non-song videos, last channel is treated as artist
    if artist_index is None and user_channel_indexes:
        artist_index = user_channel_indexes[-1]

    title = get_item_text(data, title_index) if title_index is not None else None
    if title == "Song deleted":
        return None

    artists = parse_song_artists(data, artist_index) if artist_index is not None else None

    album = parse_song_album(data, album_index) if album_index is not None else None

    views = get_item_text(data, 2) if is_album else None

    duration = get_item_text(data, duration_index) if duration_index else None
    if "fixedColumns" in data:
        if "simpleText" in nav(get_fixed_column_item(data, 0), ["text"]):
            duration = nav(get_fixed_column_item(data, 0), ["text", "simpleText"])
        else:
            duration = nav(get_fixed_column_item(data, 0), TEXT_RUN_TEXT)

    thumbnails = nav(data, THUMBNAILS, True)

    isExplicit = nav(data, BADGE_LABEL, True) is not None

    videoType = nav(
        data,
        [*MENU_ITEMS, 0, MNIR, "navigationEndpoint", *NAVIGATION_VIDEO_TYPE],
        True,
    )

    voting_status = nav(data, ENGAGEMENT_BAR, none_if_absent=True)

    community_vote_status = (
        None
        if voting_status is None
        else {
            "netVoteValue": voting_status["votes"],
            "status": VoteStatus(voting_status["status"]),
        }
    )

    song = {
        "videoId": videoId,
        "title": title,
        "artists": artists,
        "album": album,
        "likeStatus": like,
        **(song_menu_data),
        "thumbnails": thumbnails,
        "isAvailable": isAvailable,
        "isExplicit": isExplicit,
        "videoType": videoType,
        "views": views,
        "communityVoteStatus": community_vote_status,
    }

    if is_album:
        song["trackNumber"] = int(nav(data, ["index", "runs", 0, "text"])) if isAvailable else None

    if duration:
        song["duration"] = duration
        song["duration_seconds"] = parse_duration(duration)
    if setVideoId:
        song["setVideoId"] = setVideoId
    if creditsBrowseId:
        song["creditsBrowseId"] = creditsBrowseId

    return song


def validate_playlist_id(playlistId: str) -> str:
    return playlistId if not playlistId.startswith("VL") else playlistId[2:]
